
    /* 
    Copyright Emanouil Atanassov, 2003
    This file is part of the Fast Generator for the Halton Sequences.

    The Fast Generator for the Halton Sequences
    is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

     The Fast Generator for the Halton Sequences 
    is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the Fast Generator for the Halton Sequences; 
    if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "halton.h"
#include "haltondat.h"

#define MAXDATALEN 1000000
#ifdef  HALTON_P4
static halton_real  __attribute__ ((aligned (16))) databuf[MAXDATALEN];
static int     __attribute__ ((aligned (16)))  locatori[HALTON_MAXDIM];
#else
static halton_real databuf[MAXDATALEN];
static int locatori[HALTON_MAXDIM];
#endif
static int haltondim;
static int haltonissimple=0;
static int maxd(double N, int prime){
int i=0;
double p=1;
for (;p<N;p=p*prime,i++);
return i+1;
}
static double power(int prime,int stepen){
	return pow((double)1/prime,-stepen);
}
static int ostat(double N, int p){
	return (int) (N-floor(N/p)*p);
}

// GLOBAL

void haltonpreprocess(double fromN, int dim,
int modified,int**scrambles){
/* modified != 0 - modificied Halton sequence, 
modified=0 - original Halton sequence,
scrambles!=0 - with scrambling
scrambling=0 - without scrambling
*/
double pomN;
int maxdiginums[HALTON_MAXDIM];
int cifrifromN[HALTON_MAXDIM][HALTON_MAXDIGS];
int positions[HALTON_MAXDIM][HALTON_MAXDIGS];
int mnojiteli[HALTON_MAXDIM][HALTON_MAXDIGS];
halton_real dobaviteli[HALTON_MAXDIM][HALTON_MAXDIGS];
halton_real izvaditeli[HALTON_MAXDIM][HALTON_MAXDIGS];
halton_real chastichni[HALTON_MAXDIM][HALTON_MAXDIGS];
halton_real *bufptr;
int *kmodptr;
int pos;
short*shortptr ;
int*intptr;
int i,j,k;
double maxN=pow(2,50)-1;
/* maximum 2^50 numbers */
if (modified){
	kmodptr=haltonkmod;
}else{
	kmodptr=haltonkmod1;
}
haltondim=dim;
//presmetni pri koia osnova kolko maximalno cifri
for (i=0;i<dim;i++){
	maxdiginums[i]=maxd( maxN,haltonprimes[i])+1;
}
//presmetni cifrite na fromN,
for (i=0;i<dim;i++){
	for (j=0;j<HALTON_MAXDIGS;j++){
		cifrifromN[i][j]=0;
		positions[i][j]=0;
		dobaviteli[i][j]=0;
		chastichni[i][j]=0;
	}
}
for (i=0;i<dim;i++){
	pomN=fromN;
	for (j=0;j<maxdiginums[i];j++){
		cifrifromN[i][j]=ostat(pomN,haltonprimes[i]);
		pomN=floor(pomN/haltonprimes[i]);
		if (j==0)
		mnojiteli[i][0]=kmodptr[i];
		else mnojiteli[i][j]=(mnojiteli[i][j-1]*kmodptr[i])%haltonprimes[i];

		dobaviteli[i][j]=mnojiteli[i][j]*power(haltonprimes[i],-j-1);
		izvaditeli[i][j]=power(haltonprimes[i],-j);
		if (j==maxdiginums[i]-1){
			positions[i][j]=0;
		}else{
		positions[i][j]=(((scrambles)?scrambles[i][j]:0)+
			cifrifromN[i][j]*mnojiteli[i][j])%haltonprimes[i];
		}
	}	
} 
//presmetni chastichnite sumi
for (i=0;i<dim;i++){
	for(j=0;j<=maxdiginums[i]-1;j++){
		pomN=0;
		for (k=0;k<j;k++){
			pomN+=((scrambles)?scrambles[i][k]:0)*power(haltonprimes[i],-k-1);  
		}
		 for (k=j;k<maxdiginums[i];k++){ 
			pomN+=positions[i][k]*power(haltonprimes[i],-k-1);
		} 
		chastichni[i][j]=pomN;
	}
}
//zapochni zapalvane na bufera
#ifdef HALTON_P4 
	#ifdef HALTON_SINGLE
bufptr=databuf;
#define howm 8
#define howmp2 3 
for(i=0;i<(dim>>howmp2);i++){
	for(j=0;j<howm;j++){ 
		bufptr[j]=chastichni[i*howm+j][0];
	}
	bufptr+=howm;
	for(j=0;j<howm;j++){ 
		bufptr[j]=dobaviteli[i*howm+j][0];
	}
	// niama izvaditeli  tyka!
	bufptr+=howm;
	shortptr=(short*)(bufptr);
        bufptr+=2*howm;       
	for(j=0;j<howm;j++){
		shortptr[j]=positions[i*howm+j][0]; 
	}
	shortptr+=howm;
	for(j=0;j<howm;j++){
		shortptr[j]=mnojiteli[i*howm+j][0]; 
	}
	shortptr+=howm;
        for(j=0;j<howm;j++){
                shortptr[j]=haltonprimes[i*howm+j];
        }	
	shortptr+=howm;      
	for(j=0;j<howm;j++){
                shortptr[j]=((scrambles)?scrambles[i*howm+j][0]:0);
        }

}
if (dim&7){
for(j=0;howm*i+j<dim;j++){       
	bufptr[j]=1.+(float)chastichni[i*howm+j][0];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){   
	bufptr[j]=dobaviteli[i*howm+j][0];          
}
bufptr+=8;
shortptr=(short*)bufptr;         
for(j=0;howm*i+j<dim;j++){
	shortptr[j]=positions[i*howm+j][0]; 
}
shortptr+=8;
for(j=0;howm*i+j<dim;j++){  
	shortptr[j]=mnojiteli[i*howm+j][0]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){ 
	shortptr[j]=haltonprimes[i*howm+j]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){       
	shortptr[j]=((scrambles)?scrambles[i*howm+j][0]:0); 
}
bufptr+=16;
}
for(i=0;i<(dim>>howmp2);i++){
	for(j=0;j<howm;j++){ 
		bufptr[j]=chastichni[i*howm+j][1];
	}
	bufptr+=howm;
	for(j=0;j<howm;j++){ 
		bufptr[j]=dobaviteli[i*howm+j][1];
	}
	bufptr+=howm;
        for(j=0;j<howm;j++){
		bufptr[j]=izvaditeli[i*howm+j][1];
	}
	bufptr+=howm;
	shortptr=(short*)(bufptr);
        bufptr+=2*howm;       
	for(j=0;j<howm;j++){
		shortptr[j]=positions[i*howm+j][1]; 
	}
	shortptr+=howm;
	for(j=0;j<howm;j++){
		shortptr[j]=mnojiteli[i*howm+j][1]; 
	}
	shortptr+=howm;
        for(j=0;j<howm;j++){
                shortptr[j]=haltonprimes[i*howm+j];
        }	
	shortptr+=howm;      
	for(j=0;j<howm;j++){
                shortptr[j]=(scrambles)?scrambles[i*howm+j][1]:0;
        }

}
if (dim&7){
for(j=0;howm*i+j<dim;j++){       
	bufptr[j]=chastichni[i*howm+j][1];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){   
	bufptr[j]=dobaviteli[i*howm+j][1];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){
	bufptr[j]=izvaditeli[i*howm+j][1];          
}
bufptr+=8;
shortptr=(short*)bufptr;         
for(j=0;howm*i+j<dim;j++){
	shortptr[j]=positions[i*howm+j][1]; 
}
shortptr+=8;
for(j=0;howm*i+j<dim;j++){  
	shortptr[j]=mnojiteli[i*howm+j][1]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){ 
	shortptr[j]=haltonprimes[i*howm+j]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){       
	shortptr[j]=(scrambles)?scrambles[i*howm+j][1]:0; 
}
bufptr+=16;
}
//zapalni s ostatachnata informacia
//zapalni s informacia koe kade pochva - triabva da go 
//presmetnem
pos=0;
intptr=(int*)bufptr;
for(i=0;i<dim;i++){
	intptr[i]=(int)(bufptr+dim+pos);//tyka niama izlishni bytes
	pos=pos+(maxdiginums[i]-2>=0?(maxdiginums[i]-2):0)*5;
}
bufptr=bufptr+dim;
for(i=0;i<dim;i++){
	for (j=2;j<maxdiginums[i];j++){
	        shortptr=(short*)(bufptr+3);
		bufptr[0]=chastichni[i][j];
       		bufptr[1]=dobaviteli[i][j];
		bufptr[2]=izvaditeli[i][j]; 
		shortptr[0]=positions[i][j];
		shortptr[1]=mnojiteli[i][j];
		shortptr[2]=haltonprimes[i];
		shortptr[3]=(scrambles)?scrambles[i][j]:0;
		bufptr+=5;
	}
}	
	#else
	bufptr=databuf;
	#define howm 8
	#define howmp2 3 
for(i=0;i<(dim>>howmp2);i++){
	for(j=0;j<howm;j++){ 
		bufptr[j]=chastichni[i*howm+j][0];
	}
	bufptr+=howm;
	for(j=0;j<howm;j++){ 
		bufptr[j]=dobaviteli[i*howm+j][0];
	}
	bufptr+=howm;
        for(j=0;j<howm;j++){
		bufptr[j]=izvaditeli[i*howm+j][0];
	}
	bufptr+=howm;
	shortptr=(short*)(bufptr);
        bufptr+=howm;       
	for(j=0;j<howm;j++){
		shortptr[j]=positions[i*howm+j][0]; 
	}
	shortptr+=howm;
	for(j=0;j<howm;j++){
		shortptr[j]=mnojiteli[i*howm+j][0]; 
	}
	shortptr+=howm;
        for(j=0;j<howm;j++){
                shortptr[j]=haltonprimes[i*howm+j];
        }	
	shortptr+=howm;      
	for(j=0;j<howm;j++){
                shortptr[j]=(scrambles)?scrambles[i*howm+j][0]:0;
        }

}
if (dim&7){
for(j=0;howm*i+j<dim;j++){       
	bufptr[j]=chastichni[i*howm+j][0];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){   
	bufptr[j]=dobaviteli[i*howm+j][0];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){
	bufptr[j]=izvaditeli[i*howm+j][0];          
}
bufptr+=8;
shortptr=(short*)bufptr;         
for(j=0;howm*i+j<dim;j++){
	shortptr[j]=positions[i*howm+j][0]; 
}
shortptr+=8;
for(j=0;howm*i+j<dim;j++){  
	shortptr[j]=mnojiteli[i*howm+j][0]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){ 
	shortptr[j]=haltonprimes[i*howm+j]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){       
	shortptr[j]=(scrambles)?scrambles[i*howm+j][0]:0; 
}
bufptr+=8;
}
for(i=0;i<(dim>>howmp2);i++){
	for(j=0;j<howm;j++){ 
		bufptr[j]=chastichni[i*howm+j][1];
	}
	bufptr+=howm;
	for(j=0;j<howm;j++){ 
		bufptr[j]=dobaviteli[i*howm+j][1];
	}
	bufptr+=howm;
        for(j=0;j<howm;j++){
		bufptr[j]=izvaditeli[i*howm+j][1];
	}
	bufptr+=howm;
	shortptr=(short*)(bufptr);
        bufptr+=howm;       
	for(j=0;j<howm;j++){
		shortptr[j]=positions[i*howm+j][1]; 
	}
	shortptr+=howm;
	for(j=0;j<howm;j++){
		shortptr[j]=mnojiteli[i*howm+j][1]; 
	}
	shortptr+=howm;
        for(j=0;j<howm;j++){
                shortptr[j]=haltonprimes[i*howm+j];
        }	
	shortptr+=howm;      
	for(j=0;j<howm;j++){
                shortptr[j]=(scrambles)?scrambles[i*howm+j][1]:0;
        }

}
if (dim&7){
for(j=0;howm*i+j<dim;j++){       
	bufptr[j]=chastichni[i*howm+j][1];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){   
	bufptr[j]=dobaviteli[i*howm+j][1];          
}
bufptr+=8;
for(j=0;howm*i+j<dim;j++){
	bufptr[j]=izvaditeli[i*howm+j][1];          
}
bufptr+=8;
shortptr=(short*)bufptr;         
for(j=0;howm*i+j<dim;j++){
	shortptr[j]=positions[i*howm+j][1]; 
}
shortptr+=8;
for(j=0;howm*i+j<dim;j++){  
	shortptr[j]=mnojiteli[i*howm+j][1]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){ 
	shortptr[j]=haltonprimes[i*howm+j]; 
}
shortptr+=8;     
for(j=0;howm*i+j<dim;j++){       
	shortptr[j]=(scrambles)?scrambles[i*howm+j][1]:0; 
}
bufptr+=8;
}
//zapalni s ostatachnata informacia
//zapalni s informacia koe kade pochva - triabva da go 
//presmetnem
pos=0;
intptr=(int*)bufptr;
for(i=0;i<dim;i++){
	intptr[i]=(int)(bufptr+dim+pos);//tyka malko izlishni bytes, ama niama kak
	pos=pos+(maxdiginums[i]-2>=0?(maxdiginums[i]-2):0)*4;
}
bufptr=bufptr+dim;
for(i=0;i<dim;i++){
	for (j=2;j<maxdiginums[i];j++){
	        shortptr=(short*)(bufptr+3);
		bufptr[0]=chastichni[i][j];
       		bufptr[1]=dobaviteli[i][j];
		bufptr[2]=izvaditeli[i][j]; 
		shortptr[0]=positions[i][j];
		shortptr[1]=mnojiteli[i][j];
		shortptr[2]=haltonprimes[i];
		shortptr[3]=(scrambles)?scrambles[i][j]:0;
		bufptr+=4;
	}
}	
	#endif

#else
bufptr=databuf;
for(i=0;i<dim;i++){
	//zapalni cifri i pozicii
	shortptr=(short*)bufptr;
	
	shortptr[0]=positions[i][0];	
        shortptr[1]=mnojiteli[i][0];
	shortptr[2]=haltonprimes[i];
	shortptr[3]=(scrambles)?scrambles[i][0]:0;
#ifdef  HALTON_SINGLE
	bufptr[2]=chastichni[i][0];
        bufptr[3]=dobaviteli[i][0];
	bufptr+=4;
#else
	bufptr[1]=chastichni[i][0];
        bufptr[2]=dobaviteli[i][0];
	bufptr+=3;
#endif
}
for(i=0;i<dim;i++){
	shortptr=(short*)bufptr;
	
	shortptr[0]=positions[i][1];	
        shortptr[1]=mnojiteli[i][1];
	shortptr[2]=haltonprimes[i];
	shortptr[3]=(scrambles)?scrambles[i][1]:0;
#ifdef	HALTON_SINGLE
	bufptr[2]=chastichni[i][1];
        bufptr[3]=dobaviteli[i][1];
        bufptr[4]=izvaditeli[i][1];
	bufptr+=5;
#else
        bufptr[1]=chastichni[i][1];
        bufptr[2]=dobaviteli[i][1];
        bufptr[3]=izvaditeli[i][1];     
        bufptr+=4;
#endif
	
}
//zapalni s ostatachnata informacia
//zapalni s informacia koe kade pochva - triabva da go 
//presmetnem
pos=0;
intptr=(int*)bufptr;
for(i=0;i<dim;i++){
	intptr[i]=(int)(bufptr+dim+pos);
	#ifdef  HALTON_SINGLE
		pos=pos+(maxdiginums[i]-2>=0?(maxdiginums[i]-2):0)*5;
	#else
		pos=pos+(maxdiginums[i]-2>=0?(maxdiginums[i]-2):0)*4;
	#endif
}
bufptr=bufptr+dim;
for(i=0;i<dim;i++){
	for (j=2;j<maxdiginums[i];j++){
	        shortptr=(short*)bufptr;
		shortptr[0]=positions[i][j];
		shortptr[1]=mnojiteli[i][j];
		shortptr[2]=haltonprimes[i];
		shortptr[3]=(scrambles)?scrambles[i][j]:0;
#ifdef  HALTON_SINGLE
	        bufptr[2]=chastichni[i][j];
        	bufptr[3]=dobaviteli[i][j];
        	bufptr[4]=izvaditeli[i][j];
        	bufptr+=5;
#else
	
		bufptr[1]=chastichni[i][j];
       		bufptr[2]=dobaviteli[i][j];
		bufptr[3]=izvaditeli[i][j]; 
		bufptr+=4;
#endif
	}
}	
#endif
}

void haltonnextpoint(halton_real*points){
#if HALTON_P4
#ifdef HALTON_SINGLE
#define regcopy(num1,num2) __asm__ __volatile__ ("movaps %xmm" #num1 ",%xmm" #num2   )
#define loads(mem,num1) __asm__ __volatile__ ("movaps %0,%%xmm" #num1 : :"m"(mem)   )
#define loadshorts(mem,num1) __asm__ __volatile__ ("movaps %0,%%xmm" #num1 : :"m"(mem)   )
#define addshorts(mem,num1) __asm__ __volatile__ ("paddw %0,%%xmm" #num1 : :"m"(mem)   )
#define subshorts(mem,num1) __asm__ __volatile__ ("psubw %0,%%xmm" #num1 : :"m"(mem)   )
#define storess(num1,mem) __asm__ __volatile__ ("movaps %%xmm" #num1 ",%0" : :"m"(mem)   )
#define storeshort(num1,mem) __asm__ __volatile__ ("movaps %%xmm" #num1 ",%0" : :"m"(mem)   )
#define adddd(mem,num1) __asm__ __volatile__ ("addpd %0,%%xmm" #num1 : :"m"(mem)   ) 
#define addss(mem,num1) __asm__ __volatile__ ("addps %0,%%xmm" #num1 : :"m"(mem)   ) 
#define andnotmem(mem,num1) __asm__ __volatile__ ("pandn %0,%%xmm" #num1 : :"m"(mem)   ) 
#define subss(num1,num2) __asm__ __volatile__ ("subps   %xmm" #num1 ",%xmm" #num2   )  
#define addsreg(num1,num2) __asm__ __volatile__ ("addps %xmm" #num1 ",%xmm" #num2   )  
#define subdd(num1, num2) __asm__ __volatile__ ("subpd %xmm" #num1 ",%xmm" #num2 )   
#define anddd(mem,num1) __asm__ __volatile__ ("andpd %0,%%xmm" #num1 : :"m"(mem)   ) 
#define anddn(mem,num1) __asm__ __volatile__ ("andnps %0,%%xmm" #num1 : :"m"(mem)   ) 
#define setz(num1) __asm__ __volatile__ ("pxor %xmm" #num1 ",%xmm" #num1 )//  - greshno
#define cmpgtsigned(num1,num2) __asm__ __volatile__ ("pcmpgtw %xmm" #num1 ",%xmm" #num2 ) 
#define cmpeqwords(mem,num1) __asm__ __volatile__ ("pcmpeqw %0,%%xmm" #num1  : :"m"(mem)    ) 
#define selmin(num1,num2) __asm__ __volatile__ ("pminsw %xmm" #num1 ",%xmm" #num2 ) 
#define unpackhw(num1,num2) __asm__ __volatile__ ("punpckhwd %xmm" #num1 ",%xmm" #num2 ) 
#define unpacklw(num1,num2) __asm__ __volatile__ ("punpcklwd %xmm" #num1 ",%xmm" #num2 ) 
#define unpackhd(num1,num2) __asm__ __volatile__ ("punpckhdq %xmm" #num1 ",%xmm" #num2 ) 
#define unpackld(num1,num2) __asm__ __volatile__ ("punpckldq %xmm" #num1 ",%xmm" #num2 ) 
#define andreg(num1,num2) __asm__ __volatile__ ("pand %xmm" #num1 ",%xmm" #num2 ) 
#define orss(num1,num2) __asm__ __volatile__ ("orps %xmm" #num1 ",%xmm" #num2 ) 
#define andnreg(num1,num2) __asm__ __volatile__ ("pandn %xmm" #num1 ",%xmm" #num2 ) 
#define subreg(num1,num2) __asm__ __volatile__ ("psubw %xmm" #num1 ",%xmm" #num2 ) 
#define addreg(num1,num2) __asm__ __volatile__ ("paddw %xmm" #num1 ",%xmm" #num2 ) 
#define andnot(num1,num2) __asm__ __volatile__ ("pandn %xmm" #num1 ",%xmm" #num2 ) 
#define movmaskint(num1,reg32) __asm__ __volatile__ ("pmovmskb %%xmm" #num1 ",%0":"=r"(reg32))
#define extractw(num1,imm8,reg32)  __asm__ __volatile__ ("pextrw $" #imm8 ",%%xmm" #num1 ",%0 ":"=r"(reg32)) 
static int NN=0;
static short   __attribute__ ((aligned (16))) onesfloat[8]={0x3f80,0x3f80,0x3f80,0x3f80,
0x3f80,0x3f80,0x3f80,0x3f80};
int reg32;
int i,j,k;
int prnum;
int pos;
float *pptr;
register float* bptr;
float *ratabuf;
int ros;
short *shortptr=(short*)(databuf+howm*2);
int locpos=0,locpos1;
short prim;
int dimhowm;
NN++;
pptr=points;
/* triabva chastichnite sumi da varnem v points, 
posle da im dobavim po neshto, */
dimhowm=haltondim>>howmp2;
bptr=databuf;
locpos=0;
prnum=0;
loads(onesfloat[0],7);
for (i=0;i<(haltondim&~7);i+=8){
	loads(bptr[0],0);
	loads(bptr[4],1);  
	storess(0,points[0]);
        storess(1,points[4]);
	addss(bptr[howm],0);
	addss(bptr[howm+4],1);
// i//triabva da my se setne neshtoto
	
	loadshorts(shortptr[0],4);
	addshorts (shortptr[howm+0],4);
	loadshorts(shortptr[(2*howm)+0],6);
	regcopy(6,5);
       	cmpgtsigned(4,6);//   bash taka
	subreg(5,4);
        andreg(6,5);
	addreg(5,4);
		storeshort(4,shortptr[0]);
		
		cmpeqwords(shortptr[(3*howm)],4);
		movmaskint(4,reg32);
		if (reg32)
		{// v 7 sa primes
			locatori[locpos]=prnum;
			locpos+=reg32&0x01;
			locatori[locpos]=prnum+1;	
			locpos+=(reg32&0x04)>>2;
			locatori[locpos]=prnum+2;
			locpos+=(reg32&0x010)>>4;
			locatori[locpos]=prnum+3;	
			locpos+=(reg32&0x040)>>6;
			locatori[locpos]=prnum+4;
			locpos+=(reg32&0x0100)>>8;
			locatori[locpos]=prnum+5;
			locpos+=(reg32&0x0400)>>10;
			locatori[locpos]=prnum+6; 
			locpos+=(reg32&0x1000)>>12;
			locatori[locpos]=prnum+7;
			locpos+=(reg32&0x4000)>>14;
		}	
		prnum+=8;
		andnreg(7,6);
		setz(5);
                regcopy(5,4);
                unpacklw(6,5);
                subss(5,0);
                storess(0,bptr[0]);
                unpackhw(6,4);
                subss(4,1);
                storess(1,bptr[4]);



		points+=8;
		shortptr+=8*4*2;
		bptr+=8*4;
}
/*********** dosta provereno! 
* ******************* */

if (haltondim&4){
        loads(bptr[0],0);
        storess(0,points[0]);
        addss(bptr[howm],0);
// i//triabva da my se setne neshtoto
 
        loadshorts(shortptr[0],4);
        addshorts (shortptr[howm+0],4);
        loadshorts(shortptr[(2*howm)+0],6);
        regcopy(6,5);
        cmpgtsigned(4,6);//   bash taka
        subreg(5,4);
        andreg(6,5);
        addreg(5,4);
	__asm__ __volatile__ ("movq  %%xmm4,  %0"::"m"(shortptr[0]));    
        cmpeqwords(shortptr[(3*howm)],4);
        movmaskint(4,reg32);
	if (reg32)
                {// v 7 sa primes
                        locatori[locpos]=prnum;
                        locpos+=reg32&0x01;
                        locatori[locpos]=prnum+1;
                        locpos+=(reg32&0x04)>>2;
                        locatori[locpos]=prnum+2;
                        locpos+=(reg32&0x010)>>4;
                        locatori[locpos]=prnum+3;
                        locpos+=(reg32&0x040)>>6;
		}
	prnum+=4;
	andnreg(7,6);
        setz(5);
        unpacklw(6,5);
        subss(5,0);
        storess(0,bptr[0]);
	points+=4;
        shortptr+=4;
        bptr+=4;
}
if (haltondim&2){
        points[0]=bptr[0];
	points[1]=bptr[1];
	bptr[0]+=bptr[howm];	
	bptr[1]+=bptr[1+howm];	
	shortptr[0]+=shortptr[howm];
	shortptr[1]+=shortptr[howm+1];
 
	if (shortptr[0]>=shortptr[2*howm]){
                        shortptr[0]-=shortptr[2*howm];
			bptr[0]-=bptr[2*howm];
        }
        if ((shortptr[0]==shortptr[3*howm])){
                        locatori[locpos++]=prnum;
        }	
	if (shortptr[1]>=shortptr[1+2*howm]){
			bptr[1]-=bptr[1+2*howm];
                        shortptr[1]-=shortptr[1+2*howm];
        }
	if (shortptr[1]==shortptr[3*howm+1]){
                        locatori[locpos++]=prnum+1;
        }
	prnum+=2;

	points+=2;
        shortptr+=2;
        bptr+=2;
}

if (haltondim&1){
	points[0]=bptr[0]-1;
	shortptr[0]+=shortptr[howm];
	bptr[0]+=bptr[howm];
	if (shortptr[0]>=shortptr[2*howm]){
                        shortptr[0]-=shortptr[2*howm];
			bptr[0]-=bptr[2*howm];
        }

	if ((shortptr[0]==shortptr[3*howm])){
                        locatori[locpos++]=prnum;
        }

}
/* finished the unrolled loop */
if (locpos){
	locpos1=locpos;
	locpos=0;
ratabuf=databuf+(((haltondim&(~7))+(((haltondim&7)!=0)<<3))<<2);

	for(i=0;i<locpos1;i++){
		pos=locatori[i];
		if (pos<8) {
			ros=pos ; 
			bptr=ratabuf+pos;     
			shortptr=(short*)(ratabuf+3*howm)+pos;
		}else {
			ros=((pos&~7)<<2)+(pos&7);
			bptr=ratabuf+ros+(pos&~7);     
			shortptr=(short*)(bptr-(ros&7)+3*howm)+(ros&7);
		}
		shortptr[0]+=shortptr[howm];
		bptr[0]+=bptr[howm];
	        if (shortptr[0]>=shortptr[2*howm]){
			shortptr[0]-=shortptr[2*howm];
			bptr[0]-=bptr[2*howm];
		}
	        if ((shortptr[0]==shortptr[3*howm])){
        	                locatori[locpos++]=pos;
		}
		databuf[ros]=bptr[0];
        }
if (locpos){
		locpos1=locpos;
	        locpos=0;	
		for(i=0;i<locpos1;i++){
			pos=locatori[i];
			prim=haltonprimes[pos];
		bptr=(float*)(*((int*)(databuf+9*((haltondim&(~7))+(((haltondim&7)!=0)<<3)))+pos));
			shortptr=(short*)(bptr+3);
			j=0;
			while (1){
				shortptr[0]+=shortptr[1];
				bptr[0]+=bptr[1];
				if (shortptr[0]>=prim){ 
					bptr[0]-=bptr[2];
					shortptr[0]-=prim;
				}
				if  (shortptr[0]!=shortptr[3]){
						break;
				}
				shortptr+=10;
				bptr+=5;
				j++;
			}
			for(k=j;k>0;k--){
				bptr[-5*k]=bptr[0];
			}
// tova sa parvoto i vtoroto - 
			if (pos<7) ros=pos ; else ros=((pos&~7)<<2)+(pos&7);  	
			ratabuf[ros+(pos&~7)]=bptr[0];
			databuf[ros]=bptr[0];
		}
	}
		
}
#else
#define regcopy(num1,num2) __asm__ __volatile__ ("movapd %xmm" #num1 ",%xmm" #num2   )
#define loadd(mem,num1) __asm__ __volatile__ ("movapd %0,%%xmm" #num1 : :"m"(mem)   )
#define loadshorts(mem,num1) __asm__ __volatile__ ("movapd %0,%%xmm" #num1 : :"m"(mem)   )
#define addshorts(mem,num1) __asm__ __volatile__ ("paddw %0,%%xmm" #num1 : :"m"(mem)   )
#define subshorts(mem,num1) __asm__ __volatile__ ("psubw %0,%%xmm" #num1 : :"m"(mem)   )
#define storedd(num1,mem) __asm__ __volatile__ ("movapd %%xmm" #num1 ",%0" : :"m"(mem)   )
#define storeshort(num1,mem) __asm__ __volatile__ ("movapd %%xmm" #num1 ",%0" : :"m"(mem)   )
#define adddd(mem,num1) __asm__ __volatile__ ("addpd %0,%%xmm" #num1 : :"m"(mem)   )
#define subdd(num1, num2) __asm__ __volatile__ ("subpd %xmm" #num1 ",%xmm" #num2 )
#define anddd(mem,num1) __asm__ __volatile__ ("andpd %0,%%xmm" #num1 : :"m"(mem)   )
#define anddn(mem,num1) __asm__ __volatile__ ("andnps %0,%%xmm" #num1 : :"m"(mem)   )
#define setz(num1) __asm__ __volatile__ ("pxor %xmm" #num1 ",%xmm" #num1 )//  - greshno
#define cmpgtsigned(num1,num2) __asm__ __volatile__ ("pcmpgtw %xmm" #num1 ",%xmm" #num2 )
#define cmpeqwords(mem,num1) __asm__ __volatile__ ("pcmpeqw %0,%%xmm" #num1  : :"m"(mem)    )
#define selmin(num1,num2) __asm__ __volatile__ ("pminsw %xmm" #num1 ",%xmm" #num2 )
#define unpackhw(num1,num2) __asm__ __volatile__ ("punpckhwd %xmm" #num1 ",%xmm" #num2 )
#define unpacklw(num1,num2) __asm__ __volatile__ ("punpcklwd %xmm" #num1 ",%xmm" #num2 )
#define unpackhd(num1,num2) __asm__ __volatile__ ("punpckhdq %xmm" #num1 ",%xmm" #num2 )
#define unpackld(num1,num2) __asm__ __volatile__ ("punpckldq %xmm" #num1 ",%xmm" #num2 )
#define andreg(num1,num2) __asm__ __volatile__ ("pand %xmm" #num1 ",%xmm" #num2 )
#define andnreg(num1,num2) __asm__ __volatile__ ("pandn %xmm" #num1 ",%xmm" #num2 )
#define subreg(num1,num2) __asm__ __volatile__ ("psubw %xmm" #num1 ",%xmm" #num2 )
#define addreg(num1,num2) __asm__ __volatile__ ("paddw %xmm" #num1 ",%xmm" #num2 )
#define andnot(num1,num2) __asm__ __volatile__ ("pandn %xmm" #num1 ",%xmm" #num2 )
#define movmaskint(num1,reg32) __asm__ __volatile__ ("pmovmskb %%xmm" #num1 ",%0":"=r"(reg32))
#define extractw(num1,imm8,reg32)  __asm__ __volatile__ ("pextrw $" #imm8 ",%%xmm" #num1 ",%0 ":"=r"(reg32))
static int NN=0;
static short   __attribute__ ((aligned (16))) oners[8]={0x3ff0, 0x3ff0,0x3ff0,0x3ff0,0x3ff0,0x3ff0,
0x3ff0,0x3ff0} ;
int reg32;
int i,j,k;
int prnum;
int pos;
register double* bptr;
double *ratabuf;
int ros;
short *shortptr=(short*)(databuf+howm*3);
int locpos=0,locpos1;
short prim;
int dimhowm;
NN++;
/* triabva chastichnite sumi da varnem v points, 
posle da im dobavim po neshto, */
dimhowm=haltondim>>howmp2;
bptr=databuf;
//setz(7);
loadshorts(oners[0],7);   
locpos=0;
prnum=0;
for (i=0;i<(haltondim&~7);i+=8){
	loadd(bptr[0],0);
	loadd(bptr[2],1);
	loadd(bptr[4],2);
	loadd(bptr[6],3);
	storedd(0,points[0]);
	storedd(1,points[2]);
	storedd(2,points[4]);
	storedd(3,points[6]);
	adddd(bptr[howm],0);
	adddd(bptr[howm+2],1);
	adddd(bptr[howm+4],2);
	adddd(bptr[howm+6],3);

	loadshorts(shortptr[0],4);
	addshorts (shortptr[howm+0],4);
	loadshorts(shortptr[(2*howm)+0],6);
	regcopy(6,5);
       	cmpgtsigned(4,6);//   bash taka
	subreg(5,4);
        andreg(6,5);
	addreg(5,4);
	andnreg(7,6);
		storeshort(4,shortptr[0]);
		
		cmpeqwords(shortptr[(3*howm)],4);
		movmaskint(4,reg32);
		if (reg32)
		{// v 7 sa primes
			locatori[locpos]=prnum;
			locpos+=reg32&0x01;
			locatori[locpos]=prnum+1;	
			locpos+=(reg32&0x04)>>2;
			locatori[locpos]=prnum+2;
			locpos+=(reg32&0x010)>>4;
			locatori[locpos]=prnum+3;	
			locpos+=(reg32&0x040)>>6;
			locatori[locpos]=prnum+4;
			locpos+=(reg32&0x0100)>>8;
			locatori[locpos]=prnum+5;
			locpos+=(reg32&0x0400)>>10;
			locatori[locpos]=prnum+6; 
			locpos+=(reg32&0x1000)>>12;
			locatori[locpos]=prnum+7;
			locpos+=(reg32&0x4000)>>14;
		}	
		prnum+=8;
// v 6 e tova, koeto triabva, 4 , 5  - free, 7 - occupied, unusable
		setz(5);
		regcopy(5,4);
		unpacklw(6,5);
		unpackld(5,4);
                subdd(4,0);
		storedd(0,bptr[0]); 
		setz(4);
		regcopy(4,0);
		unpackhd(5,4);   
                subdd(4,1);
		storedd(1,bptr[2]); 
		regcopy(0,5);
		regcopy(0,1);
		unpackhw(6,0);
		unpackld(0,1);
		subdd(1,2);
		storedd(2,bptr[4]); 
		unpackhd(0,5);
		subdd(5,3);
		storedd(3,bptr[6]);     
		points+=8;
		shortptr+=8*4*4;
		bptr+=8*4;
}
/*********** dosta provereno! 
* ******************* */

if (haltondim&4){
	loadd(bptr[0],0);
        loadd(bptr[2],1);
        storedd(0,points[0]);
        storedd(1,points[2]);
        adddd(bptr[howm],0);
        adddd(bptr[howm+2],1);
 
        loadshorts(shortptr[0],4);
        addshorts (shortptr[howm+0],4);
        loadshorts(shortptr[(2*howm)+0],6);
        regcopy(6,5);
        cmpgtsigned(4,6);//   bash taka
        regcopy(6,7);
        andnreg(5,6);
        subreg(6,4);
	 __asm__ __volatile__ ("movq  %%xmm4,  %0"::"m"(shortptr[0]));  
        cmpeqwords(shortptr[(3*howm)],4);
        movmaskint(4,reg32);
        if (reg32){
                        locatori[locpos]=prnum;
                        locpos+=reg32&0x01;
                        locatori[locpos]=prnum+1;
                        locpos+=(reg32&0x04)>>2;
                        locatori[locpos]=prnum+2;
                        locpos+=(reg32&0x010)>>4;
                        locatori[locpos]=prnum+3;
                        locpos+=(reg32&0x040)>>6;
	}
	prnum+=4;
	regcopy(7,5);
        regcopy(7,4);
        unpacklw(4,4);
        regcopy(4,6);
        unpacklw(4,4);
        anddn(bptr[2*howm+0],4);
 
        subdd(4,0);
        storedd(0,bptr[0]);
        unpackhw(6,6);
                // 6 ready again
        anddn(bptr[2*howm+2],6);
        subdd(6,1);
        storedd(1,bptr[2]);
	/* 
        shortptr+=4*4*4;
        bptr+=4*4;
	*/
	points+=4;
        shortptr+=4;
        bptr+=4;
}
if (haltondim&2){
/***   Ne moje! 
	loadd(bptr[0],0);
        storedd(0,points[0]);
        adddd(bptr[howm],0);
 
        loadshorts(shortptr[0],4);
        addshorts (shortptr[howm+0],4);
        loadshorts(shortptr[(2*howm)+0],6);
        regcopy(6,5);
        cmpgtsigned(4,6);//   bash taka
        regcopy(6,7);
        andnreg(5,6);
        subreg(6,4);
        __asm__ __volatile__ ("movd  %%xmm4,  %0"::"m"(shortptr[0]));
        cmpeqwords(shortptr[(3*howm)],4);
        movmaskint(4,reg32);
        if (reg32){
                        locatori[locpos]=prnum;
                        locpos+=reg32&0x01;
                        locatori[locpos]=prnum+1;
                        locpos+=(reg32&0x04)>>2;
        }
        prnum+=2;
        regcopy(7,5);
        regcopy(7,4);
        unpackl(4,4);
        regcopy(4,6);
        unpackl(4,4);
        anddn(bptr[2*howm+0],4);
        subdd(4,0);
        storedd(0,bptr[0]);
        points+=2;
        shortptr+=2;
        bptr+=2;
	Ne moje i ne triabva*/



	shortptr[0]+=shortptr[howm];
	shortptr[1]+=shortptr[howm+1];
        points[0]=bptr[0];
	points[1]=bptr[1];
        bptr[0]+=bptr[howm];
	bptr[1]+=bptr[howm+1];
 
        if (shortptr[0]>=shortptr[2*howm]){
                        bptr[0]-=bptr[2*howm];
                        shortptr[0]-=shortptr[2*howm];
        }
        if ((shortptr[0]==shortptr[3*howm])){
                        locatori[locpos++]=prnum;
        }	
 	if (shortptr[1]>=shortptr[2*howm+1]){
                        bptr[1]-=bptr[2*howm+1];
                        shortptr[1]-=shortptr[2*howm+1];
        }
	if (shortptr[1]==shortptr[3*howm+1]){
                        locatori[locpos++]=prnum+1;
        }
	prnum+=2;

	points+=2;
        shortptr+=2;
        bptr+=2;
}

if (haltondim&1){
	shortptr[0]+=shortptr[howm];
	points[0]=bptr[0];
	bptr[0]+=bptr[howm];      

	if (shortptr[0]>=shortptr[2*howm]){
                        bptr[0]-=bptr[2*howm];
                        shortptr[0]-=shortptr[2*howm];
        }

	if ((shortptr[0]==shortptr[3*howm])){
                        locatori[locpos++]=prnum;
        }

}
/* finished the unrolled loop */
if (locpos){
	locpos1=locpos;
	locpos=0;
ratabuf=databuf+(((haltondim&(~7))+(((haltondim&7)!=0)<<3))<<2);

	for(i=0;i<locpos1;i++){
		pos=locatori[i];
		if (pos<8) {
			ros=pos ; 
			bptr=ratabuf+pos;     
			shortptr=(short*)(ratabuf+3*howm)+pos;
		}else {
			ros=((pos&~7)<<2)+(pos&7);
			bptr=ratabuf+ros;     
			shortptr=(short*)(bptr-(ros&7)+3*howm)+(ros&7);
		}
		shortptr[0]+=shortptr[howm];
        	bptr[0]+=bptr[howm];
	        if (shortptr[0]>=shortptr[2*howm]){
			bptr[0]-=bptr[2*howm];   
			shortptr[0]-=shortptr[2*howm];
		}
	        if ((shortptr[0]==shortptr[3*howm])){
        	                locatori[locpos++]=pos;
		}
		databuf[ros]=bptr[0];
        }
if (locpos){
		locpos1=locpos;
	        locpos=0;	
		for(i=0;i<locpos1;i++){
			pos=locatori[i];
			prim=haltonprimes[pos];
bptr=(double*)(*((int*)(databuf+8*((haltondim&(~7))+(((haltondim&7)!=0)<<3)))+pos));
			shortptr=(short*)(bptr+3);
			j=0;
			while (1){
				shortptr[0]+=shortptr[1];
			        bptr[0]+=bptr[1];
				if (shortptr[0]>=prim){ 
					bptr[0]-=bptr[2]; 
					shortptr[0]-=prim;
				}
				if  (shortptr[0]!=shortptr[3]){
						break;
				}
				shortptr+=16;
				bptr+=4;
				j++;
			}
			for(k=j;k>0;k--){
				bptr[-4*k]=bptr[0];
			}
// tova sa parvoto i vtoroto - 
			if (pos<7) ros=pos ; else ros=((pos&~7)<<2)+(pos&7);  	
			ratabuf[ros]=bptr[0];
			databuf[ros]=bptr[0];
		}
	}
		
}

#endif
#else
static int NN=0;
int i,j,k;
int pos;
halton_real* bptr;
short *shortptr=(short*)databuf;
int locpos=0,locpos1;
short prim;
#ifdef HALTON_SINGLE
bptr=databuf+2;
#else
bptr=databuf+1;
#endif
NN++;
shortptr=(short*)databuf;
for(i=0;i<haltondim>>1;i++){
	points[0]=bptr[0];
#ifdef HALTON_SINGLE
	points[1]=bptr[4];
	bptr+=8;
#else
	points[1]=bptr[3];	
	bptr+=6;
#endif
	points+=2;
}
if (haltondim&1){
	points[0]=bptr[0];
}
#ifdef HALTON_SINGLE
bptr=databuf+2;
#else
bptr=databuf+1;
#endif
if (haltonissimple){
for(i=0;i<(haltondim&~1);i+=2){
#ifdef HALTON_SINGLE
	shortptr[0]++;
	shortptr[8]++;
	bptr[0]+=bptr[1];
	bptr[4]+=bptr[5];
	if (shortptr[0]==shortptr[2]){ shortptr[0]=0;locatori[locpos++]=i; }
	if (shortptr[8]==shortptr[10]){shortptr[8]=0;locatori[locpos++]=i+1; }
	bptr+=8;
	shortptr+=16;
#else
	shortptr[0]++;
	shortptr[12]++;
	bptr[0]+=bptr[1];
	bptr[3]+=bptr[4];
if (shortptr[0]==shortptr[2]){  shortptr[0]=0;locatori[locpos++]=i; }
if (shortptr[12]==shortptr[2]){ shortptr[12]=0;locatori[locpos++]=i+1; }
bptr+=6;
shortptr+=24;
#endif
}
if (1&haltondim){
	shortptr[0]++;
        bptr[0]+=bptr[1];
        prim=shortptr[2];
        if (shortptr[0]==shortptr[2]){
			shortptr[0]=0;
                        locatori[locpos++]=i;
        }
}
if (locpos){
	locpos1=locpos;
	locpos=0;

	for(i=0;i<locpos1;i++){
		pos=locatori[i];
#ifdef HALTON_SINGLE
		bptr=databuf+(4*(haltondim+pos))+pos+2;
#else
		bptr=databuf+(3*(haltondim+pos))+pos+1;
#endif
#ifdef HALTON_SINGLE
		shortptr=(short*)(bptr-2);
#else
		shortptr=(short*)(bptr-1);
#endif
		shortptr[0]++;
        	bptr[0]+=bptr[1];
	        if ((shortptr[0]==shortptr[2])){
				shortptr[0]=0;
        	                locatori[locpos++]=pos;
		}
#ifdef HALTON_SINGLE
		bptr[-(haltondim*4)-pos]=bptr[0];
#else
		bptr[-(haltondim*3)-pos]=bptr[0];
#endif
        }
	if (locpos){
		locpos1=locpos;
	        locpos=0;	
		for(i=0;i<locpos1;i++){
			pos=locatori[i];
			prim=haltonprimes[pos];
			#ifdef HALTON_SINGLE
			shortptr=(short*)(*((int*)(databuf+9*haltondim)+pos));
			bptr=(halton_real*)shortptr+2;
			#else
				shortptr=(short*)(*((int*)(databuf+7*haltondim)+pos));
	                        bptr=(halton_real*)shortptr+1;
			#endif
			j=0;
			while (1){
				shortptr[0]++;
			        bptr[0]+=bptr[1];
				if  (shortptr[0]!=prim){
						break;
				}
				shortptr[0]=0;
				#ifdef HALTON_SINGLE
					shortptr+=10;
                                        bptr+=5;
				#else
					shortptr+=16;
					bptr+=4;
				#endif
				j++;
			}
			for(k=j;k>0;k--){
				#ifdef HALTON_SINGLE
					bptr[-5*k]=bptr[0];
				#else
					bptr[-4*k]=bptr[0];
				#endif
			}
			#ifdef HALTON_SINGLE
				databuf[pos*4+2]=bptr[0];
				databuf[(haltondim+pos)*4 + pos+2]=bptr[0];
			#else
				databuf[pos*3+1]=bptr[0];
				databuf[(pos+haltondim)*3+pos+1]=bptr[0];
			#endif

		}
	}
		
}
}else{
for(i=0;i<(haltondim&~1);i+=2){
#ifdef HALTON_SINGLE
	shortptr[0]+=shortptr[1];
	shortptr[8]+=shortptr[9];
	bptr[0]+=bptr[1];
	bptr[4]+=bptr[5];
	if (shortptr[0]>=shortptr[2]){
                shortptr[0]-=shortptr[2];
                bptr[0]-=1.f; //moje i taka
        }
	if (shortptr[8]>=shortptr[10]){
                shortptr[8]-=shortptr[10];
                bptr[4]-=1.f; //moje i taka
        }
	if (shortptr[0]==shortptr[3]){ locatori[locpos++]=i; }
	if (shortptr[8]==shortptr[11]){ locatori[locpos++]=i+1; }
	bptr+=8;
	shortptr+=16;
#else
	shortptr[0]+=shortptr[1];
	shortptr[12]+=shortptr[13];
	bptr[0]+=bptr[1];
	bptr[3]+=bptr[4];
	if (shortptr[0]>=shortptr[2]){
		shortptr[0]-=shortptr[2];
		bptr[0]-=1.; //moje i taka
	}
	if (shortptr[12]>=shortptr[14]){
		shortptr[12]-=shortptr[14];
		bptr[3]-=1.; //moje i taka
	}
if (shortptr[0]==shortptr[3]){ locatori[locpos++]=i; }
if (shortptr[12]==shortptr[15]){ locatori[locpos++]=i+1; }
bptr+=6;
shortptr+=24;
#endif
}
if (1&haltondim){
	shortptr[0]+=shortptr[1];
        bptr[0]+=bptr[1];
        prim=shortptr[2];
        if (shortptr[0]>=prim){
                shortptr[0]-=prim;
                bptr[0]-=1.; //moje i taka
        }
        if (shortptr[0]==shortptr[3]){
                        locatori[locpos++]=i;
        }
}
if (locpos){
	locpos1=locpos;
	locpos=0;

	for(i=0;i<locpos1;i++){
		pos=locatori[i];
#ifdef HALTON_SINGLE
		bptr=databuf+(4*(haltondim+pos))+pos+2;
#else
		bptr=databuf+(3*(haltondim+pos))+pos+1;
#endif
#ifdef HALTON_SINGLE
		shortptr=(short*)(bptr-2);
#else
		shortptr=(short*)(bptr-1);
#endif
		shortptr[0]+=shortptr[1];
        	bptr[0]+=bptr[1];
	        if (shortptr[0]>=shortptr[2]){
			bptr[0]-=bptr[2];   
			shortptr[0]-=shortptr[2];
		}
	        if ((shortptr[0]==shortptr[3])){
        	                locatori[locpos++]=pos;
		}
#ifdef HALTON_SINGLE
		bptr[-(haltondim*4)-pos]=bptr[0];
#else
		bptr[-(haltondim*3)-pos]=bptr[0];
#endif
        }
	if (locpos){
		locpos1=locpos;
	        locpos=0;	
		for(i=0;i<locpos1;i++){
			pos=locatori[i];
			prim=haltonprimes[pos];
			#ifdef HALTON_SINGLE
			shortptr=(short*)(*((int*)(databuf+9*haltondim)+pos));
			bptr=(halton_real*)shortptr+2;
			#else
				shortptr=(short*)(*((int*)(databuf+7*haltondim)+pos));
	                        bptr=(halton_real*)shortptr+1;
			#endif
			j=0;
			while (1){
				shortptr[0]+=shortptr[1];
			        bptr[0]+=bptr[1];
				if (shortptr[0]>=prim){ 
					bptr[0]-=bptr[2]; 
					shortptr[0]-=prim;
				}
				if  (shortptr[0]!=shortptr[3]){
						break;
				}
				#ifdef HALTON_SINGLE
					shortptr+=10;
                                        bptr+=5;
				#else
					shortptr+=16;
					bptr+=4;
				#endif
				j++;
			}
			for(k=j;k>0;k--){
				#ifdef HALTON_SINGLE
					bptr[-5*k]=bptr[0];
				#else
					bptr[-4*k]=bptr[0];
				#endif
			}
			#ifdef HALTON_SINGLE
				databuf[pos*4+2]=bptr[0];
				databuf[(haltondim+pos)*4 + pos+2]=bptr[0];
			#else
				databuf[pos*3+1]=bptr[0];
				databuf[(pos+haltondim)*3+pos+1]=bptr[0];
			#endif

		}
	}
		
}
}
#endif
}
double halton_test_get_point(double N,int prime,int kmod,int *scrambles)
{
int digits[64];
int kmods[64];
long int i,dim,rn;
double s;
double maxN=pow(2,50)-1;
dim=maxd( maxN,prime); /* tykka triabva da e s edno po-malko, 
zashtoto otgovaria na nula */
        rn=N;
        for(i=0;i<dim;i++){
                digits[i]=rn%prime;
                rn=rn/prime;
                if (i==0){
                        kmods[0]=kmod;
                }else{
                        kmods[i]=(kmods[i-1]*kmod)%prime;
                }
        }
	
        s=0;
        for(i=dim-1;i>=0;i--){
                s+=((kmods[i]*digits[i]+((scrambles)?scrambles[i]:0))%prime)*pow(prime,-i-1);
        }
        return s;
}


static double rannum(){
	return (double)rand()/(double)RAND_MAX;
}


static int **scrambles;
// GLOBALS
halton_real *haltonptr;
halton_real *haltonrbound;
halton_real haltonpoints[HALTON_MAXDIM];
double (*haltonranfunc)();

void haltonprepare(double from,int modified, int scrambled, int dim){
int i,j;
	haltonissimple=(scrambled==0)&&(modified==0);
	 scrambles=(int**)malloc(sizeof(int*)*dim);
	if (scrambled){
         for (i=0;i<dim;i++){
                scrambles[i]=(int*)malloc(sizeof(int)*HALTON_MAXDIGS);
                for (j=0;j<HALTON_MAXDIGS;j++) scrambles[i][j]=
                        (int)(rannum()*(haltonprimes[i]));    /* between 0 and primes[i]-1;   */
         }
	}else{
                scrambles=NULL;
	}
	 haltonpreprocess(from,dim,modified,scrambles);
         haltonptr=haltonpoints;
         haltonrbound=haltonptr+dim;
}
void haltonnexttraj(){
        haltonptr=haltonpoints;
        haltonnextpoint(haltonpoints);
}

halton_real*haltonalignmem(int numreals){
        void*p;
        p=malloc(numreals*sizeof(double)+16);
        if ((int)p&~0xf){
                p=(void*)(((int)p+0x10)&~0xf);
        }
        return (halton_real*)p;
}

