
    /* 
    Copyright Emanouil Atanassov, 2003
    This file is part of the Fast Generator for the Halton Sequences.

    The Fast Generator for the Halton Sequences
    is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

     The Fast Generator for the Halton Sequences 
    is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the Fast Generator for the Halton Sequences; 
    if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include <unistd.h>
#include <sys/time.h>

#include "halton.h"


double msecs() /* milliseconds */
{
struct timeval tv;
struct timezone tz;
tz.tz_minuteswest=-120;
tz.tz_dsttime=0;
gettimeofday(&tv, &tz);
return 0.000001*((long long)tv.tv_sec*1000000+tv.tv_usec);
 
}

double randd(){
        return (double)rand()/(double)RAND_MAX;
}


double intfun(int intnum, int dimen){
  double s;
  double p;
  int j;
  double x;
  s=1.;
  switch(intnum){
  case 1:{
	for(j=0;j<dimen;j++) {
		x=haltonnextcoordr();
		s=s*(x*x*x+0.75);
  	}
	break;
	}
  case 2:{
	for(j=0;j<dimen;j++)
  	{
   		s=s*fabs(4.0*haltonnextcoordr()-2.0);
   	}
	break;
	}
  case 3:{
	 for(j=1;j<dimen+1;j++) {
		s=s*j*cos(j*haltonnextcoordr());
   	}
	break;
	}
  case 4:{
	s=0.;
  	p=-1.;
	for(j=0;j<dimen;j++) {
   		p=-p*haltonnextcoordr();
   		s=s+p;
   	}
  }
}
  	return s;
}
double exact(int intno,int dimen){
double s,p;
int j;
	if (intno<3){
		return 1;
	}else{
		if (intno==3){
			s=1.;
        		for(j=1;j<dimen+1;j++) {
                		s=s*sin(j);
        		}
        		return s;
		}else{
			p=1./pow(2,dimen);
			if (dimen%2==0){
                   		s=(1-p)/3;
        		} else {
                   		s=(1+p)/3;
        		}
        		return s;
		}
	}
}
int main(int argc,char**argv){
	int intno;
        double m1,m2;
	double val;
        int j;
	double  nterms;
	int i;
	double ctr;
        int dimension;
	int modified;
	int scrambled;
        double from=0;
	int choice;
        int s;
        double resu;
        halton_real*result;
        int **scrambles;
        m1=msecs();
        result=haltonalignmem(HALTON_MAXDIM);
        printf("Tell me what to do?\n");
        printf("1. Test output, 2. Measure speed, 3. Calculate integrals\n");
	printf("Choice: ");
	scanf("%i",&choice);

	printf("\nDimension: ");
        scanf("%i",&dimension);
        printf("\nBegining from: ");
        scanf("%lf",&from);
        printf("\nModified (1) or Original (0):");
        scanf("%i",&modified);
        printf("\nScrambled (1) or not (0):");
        scanf("%i",&scrambled);
        printf("\nNumber of terms:");
        scanf("%lf",&nterms);

	if (scrambled){
		scrambles=(int**)malloc(sizeof(int*)*dimension);
         	for (i=0;i<dimension;i++){
        	        scrambles[i]=(int*)malloc(sizeof(int)*HALTON_MAXDIGS);
                	for (j=0;j<HALTON_MAXDIGS;j++) {	scrambles[i][j]=
                        	(int)(randd()*(haltonprimes[i]));    /* between 0 and primes[i]-1;   */
			}
		}
         }else{
		scrambles=NULL;
	 }
	 switch (choice){
		case 1:{ /* Test */
			haltonpreprocess(from ,dimension,modified,scrambles);    
			for (ctr=0;ctr<nterms;ctr++){
				haltonnextpoint(result); 
				//for(s=0;s<dimension;s++){
				s=(int)(dimension*randd());
				if (modified){
					if (scrambled){
		                        	resu=halton_test_get_point(ctr+from,haltonprimes[s],haltonkmod[s],scrambles[s]);
					}else{
						resu=halton_test_get_point(ctr+from,haltonprimes[s],haltonkmod[s],0);
					}
				}else{
					if (scrambled){
                                                resu=halton_test_get_point(ctr+from,haltonprimes[s],1,scrambles[s]);
                                        }else{
                                                resu=halton_test_get_point(ctr+from,haltonprimes[s],1,0);
					}
				}
				 
                        	if (fabs(result[s]- resu)>0.001){
                                		printf("Error found  %f %i %g %g\n", ctr,s,resu, result[s]);
						abort();
                        	}
				//}
	
        		}
			printf("\nSuccess!\n");
			break;
		}
		case 2:{ /* Speed */
			m1=msecs();
			haltonpreprocess(from ,dimension,modified,scrambles);
			for (i=0;i<nterms;i++){
				haltonnextpoint(result); 
                        }
			m2=msecs();
			printf("Last coordinate of the last point %g\n",(double)result[dimension-1]);
			printf("Time in  seconds and "
			"speed in millions of coordinates per second: %.2e %.3f\n",
			m2-m1,
			dimension/(m2-m1));
			break;
		}
		case 3:{
			printf("\nIntegral N (1-4): \n");
        		scanf("%i",&intno);
			val=0;
			m1=msecs(); 
			haltonprepare(0.,modified,scrambled,384>dimension?dimension:384);     
			haltonranfunc=randd;
			for (ctr=0;ctr<nterms;ctr++){
                                haltonnexttraj();
				val+=intfun(intno,dimension);
                        }    
			m2=msecs(); 
			printf("In %f seconds, the integral with exact value "
			"%f was approximated as %f ,\n"
			"using %.0g points of the sequence, with an absolute error of %.3e.\n",
			(m2-m1),exact(intno,dimension), val/(double)nterms,
(double) nterms, fabs(val/(double)nterms - exact(intno,dimension)));
			break;
		}
	}
	return 0;
}
